/**
 *
 *		Interior Mapping with one cubemap per face
 *		Dana Goyette
 *		CPE 473, Cal Poly Spring 2010
 *		
 *		Based on interior-mapping code from http://interiormapping.oogst3d.net/
 *
 */

//textures seem to bind 1 to 1 with texture units
//uniform bool usebase;
uniform sampler2D baseTex;


varying vec3 worldNormal;

uniform vec3 camPos;
uniform vec4 lightPos;
uniform vec4 lightColor;

uniform bool useWalls;

uniform vec3 wallFreq;
uniform vec3 offset;

//uniform sampler3D wallTex;
uniform samplerCube wallTex;

void main()
{

	if (useWalls)
	{
		vec3 direction = normalize(gl_TexCoord[1].xyz - camPos.xyz); // gl_TexCoord[1] holds real-world x,y,z
	
		// position of next HIGHER wall
		vec3 walls = (floor(wallFreq*0.999*gl_TexCoord[1].xyz) + vec3(float(direction.x >= 0.0), float(direction.y >= 0.0), float(direction.z >= 0.0))) / wallFreq; 
		
	
		vec3 rayFractions = (walls - vec3(camPos)) / direction;
		float minFrac = min(min(rayFractions.x, rayFractions.y), rayFractions.z);
	
	
		vec3 hitXYZ = (camPos + minFrac*direction) - offset;
		vec4 wallTexColor = textureCube(wallTex,-hitXYZ);
		vec3 wallNormal;
		if (minFrac == rayFractions.x)
		{
			wallNormal = (direction.x>=0) ? vec3(-1,0,0) : vec3(-1,0,0);
			gl_FragColor = vec4(0.01,0,0,0);
		}
		else if (minFrac == rayFractions.y)
		{
			wallNormal = (direction.y>=0) ? vec3(0,-1,0) : vec3(0,1,0);
			gl_FragColor = vec4(0,0.01,0,0);
		}
		else
		{
			wallNormal = (direction.z>=0) ? vec3(0,0,-1) : vec3(0,0,1);
			gl_FragColor = vec4(0,0,0.01,0);
		}
	
	

	
	
		vec4 baseTexColor = texture2D(baseTex, gl_TexCoord[0].st);


		vec3 lightVec = normalize(lightPos.xyz - gl_TexCoord[1].xyz);
	
		float diffuseDot = clamp(dot(lightVec, worldNormal), 0.0, 1.0);
	

		vec4 diffuseColor = vec4( ((0.6 * diffuseDot + 0.5) * lightColor).rgb, 1.0); // hardcode diffuseness=0.6

		if (dot(-worldNormal,direction)>0) // only show inner walls when looking from outside
		{
			gl_FragColor += diffuseColor * baseTexColor.a * vec4(baseTexColor.rgb,1.0);

		

		
			//lightVec = normalize(lightPos.xyz - hitXYZ);
	
			//diffuseDot = clamp(dot(lightVec, wallNormal), 0.0, 1.0);
			//diffuseColor = vec4(((0.6 * diffuseDot + 0.5) * lightColor).rgb, 1.0); // hardcode diffuseness=0.6
			diffuseColor = vec4(1,1,1,1); // makes it nicer and brighter

			gl_FragColor += diffuseColor * (1-baseTexColor.a) * wallTexColor;
		}
		else
		{
			gl_FragColor = diffuseColor * baseTexColor.a * vec4(baseTexColor.rgb,1.0);
		}
		
	}
	else
	{
		gl_FragColor = textureCube(wallTex, vec3(-worldNormal.x,-worldNormal.y,worldNormal.z));
		gl_FragColor.a=1;
	}

}

