/**
 *
 *		Interior Mapping with one cubemap per face
 *		Dana Goyette
 *		CPE 473, Cal Poly Spring 2010
 *
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <map>

#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#include <OPENGL/glext.h>
#endif

#include "GLSL_helper.h"
#include "tex.h"



#define N_CUBES 7
#define EXTERIOR_NAME "exterior"

std::string wallnames[N_CUBES] = {"right","left","top","bottom","front","back","world"};
//std::string wallnames[N_CUBES] = {"MarriottMadisonWest","Interior","JustWindows","Earth","DallasW","Skansen2","Pond"};
std::string wall_axes[6] = {"posx","negx", "negy", "posy","posz","negz"}; // negative and positive y are swapped in the cubemaps I'm using

const GLfloat neg=0;
const GLfloat Pos=1;

#define negXverts {{neg,neg,neg},\
				   {neg,neg,Pos},\
				   {neg,Pos,Pos},\
				   {neg,Pos,neg}}
								 
#define PosXverts {{Pos,neg,Pos},\
				   {Pos,neg,neg},\
				   {Pos,Pos,neg},\
				   {Pos,Pos,Pos}}
								 
#define negYverts {{neg,neg,neg},\
				   {Pos,neg,neg},\
				   {Pos,neg,Pos},\
				   {neg,neg,Pos}}

#define PosYverts {{neg,Pos,Pos},\
				   {Pos,Pos,Pos},\
				   {Pos,Pos,neg},\
				   {neg,Pos,neg}}

#define negZverts {{Pos,neg,neg},\
				   {neg,neg,neg},\
				   {neg,Pos,neg},\
				   {Pos,Pos,neg}}

#define PosZverts {{neg,neg,Pos},\
				   {Pos,neg,Pos},\
				   {Pos,Pos,Pos},\
				   {neg,Pos,Pos}}

const GLfloat cubeVerts[6][4][3] = {PosXverts, negXverts,
									PosYverts, negYverts,
									PosZverts, negZverts};

const GLfloat cubeNormals[6][3] = {{1,0,0},{-1,0,0},
								   {0,1,0},{0,-1,0},
								   {0,0,1},{0,0,-1}};

const GLfloat texCoords[4][2] = {{0,0},
								 {1,0},
								 {1,1},
								 {0,1}};



/* Vector3D struct mostly used only for camera nav */
typedef struct Vector3 {
  float x;
  float y;
  float z;
  
  float nx;
  float ny;
  float nz;
  Vector3(float in_x, float in_y, float in_z) : x(in_x), y(in_y), z(in_z) {}
  Vector3() : x(0), y(0), z(0) {}
  Vector3(const Vector3& in_v) : x(in_v.x), y(in_v.y), z(in_v.z) {}
  Vector3(float in) : x(in), y(in), z(in) {}
  inline Vector3& operator =(const Vector3& v2) { x = v2.x; y=v2.y; z=v2.z; return *this;}
  inline void set(float in_x, float in_y, float in_z) { x = in_x; y = in_y; z = in_z;}
} Vector3;

//globals used mostly for camera
Vector3 eye, la, up;
int GW;
int GH;
int mouseX;
int mouseY;
int startx, starty, endx, endy;
float phi, theta;

/*Some multi-texturing stuff */
int  gbMultitexturing = 1;
GLfloat gOffset = 0.0f;

GLfloat lightPos[] = {15, 12, -20};

/*Global to access the shader program - used in keyboard to turn on and off */
int ShadeProg;

void display();

GLuint* baseTexID;
TGA** baseTex;

GLuint* cubeTexID;
TGA*** cubeTex;

// http://www.programmersheaven.com/mb/beginnercpp/367089/367089/discovering-original-binary-path/

char* findMyPath(const char* argv)
{	
	static char path[300];
	int i;
	
	strncpy(path, argv, sizeof(path));
	
	for (i = strlen(path) - 1; i >= 0; --i)
	{
		if (path[i] == '\\' || path[i] == '/')
		{
			path[i + 1] = 0;
			break;
		}
	}	
	
	if (i == -1) // user just typed the binary name
		strcpy(path, ".");
				
	return path;
}

/* Code to set up using the GLSL shader */
bool InstallShader(const GLchar *vShaderText, const GLchar *fShaderText) {
	GLuint VertVS, FragFS; //handles to shader object
	GLint vCompiled, fCompiled, linked; //status of shader

	
	VertVS = glCreateShader(GL_VERTEX_SHADER);
	FragFS = glCreateShader(GL_FRAGMENT_SHADER);
	
	//load the source
	glShaderSource(VertVS, 1, &vShaderText, NULL);
	glShaderSource(FragFS, 1, &fShaderText, NULL);
	cerr<<"Vertex Shader{";
	//compile vertex shader and print log
	glCompileShader(VertVS);
	/* check shader status requires helper functions */
	printOpenGLError();
	glGetShaderiv(VertVS, GL_COMPILE_STATUS, &vCompiled);
	printShaderInfoLog(VertVS);
	cerr << '}'<< endl;
	cerr<<"Fragment Shader{";
	//compile shader and print log
	glCompileShader(FragFS);
	/* check shader status requires helper functions */
	printOpenGLError();
	glGetShaderiv(FragFS, GL_COMPILE_STATUS, &fCompiled);
	printShaderInfoLog(FragFS);
	cerr << '}' << endl;
	if (!vCompiled || !fCompiled) {
		printf("Error compiling the shader -- vertex: %s; frag: %s.", vCompiled ? "succeeded" : "failed", fCompiled ? "succeeded" : "failed");
		return false;
	}
	
	//create a program object and attach the compiled shader
	ShadeProg = glCreateProgram();
	glAttachShader(ShadeProg, VertVS);
	glAttachShader(ShadeProg, FragFS);
	
	glLinkProgram(ShadeProg);
	/* check shader status requires helper functions */
	printOpenGLError();
	glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
	printProgramInfoLog(ShadeProg);

	printf("sucessfully installed shader\n");
	return true;
	
}	
										
/* Initialization for some OpenGL stuff */
bool Initialize ()					
{

	// Start Of User Initialization
	glClearColor (0.0f, 0.0f, 0.0f, 0.5f);								// Black Background
	glClearDepth (1.0f);												// Depth Buffer Setup
	glDepthFunc (GL_LEQUAL);											// The Type Of Depth Testing
	glEnable (GL_DEPTH_TEST);											// Enable Depth Testing
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);					// Set Perspective Calculations To Most Accurate
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);							// Draw polygons filled in

	return true;														// Return true (Initialization Successful)
}


//helper functions for working with vectors
Vector3 cross(const Vector3& a, const Vector3& b) {

	Vector3 c;
	c.x = a.y*b.z - a.z*b.y;
	c.y = a.z*b.x - a.x*b.z;
	c.z = a.x*b.y - a.y*b.x;

	return(c);
}

float dot(const Vector3& a, const Vector3& b) {
	return(a.x*b.x + a.y*b.y + a.z*b.z);
}

float length(const Vector3& a) {
	return(sqrt(dot(a, a)));
}

Vector3 normalize(const Vector3& a) {
	Vector3 u;
	float mag_a = length(a);
	u.x = a.x/mag_a;
	u.y = a.y/mag_a;
	u.z = a.z/mag_a;
	return(u);
}


/*Draw a cube face by face */
void DrawCube() {

	glEnable(GL_BLEND);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


	glEnable(GL_TEXTURE_2D);



	glEnable(GL_TEXTURE_CUBE_MAP);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexID[N_CUBES-1]); // each face gets its own set of cubemap textures
	
	glUseProgram(ShadeProg);
	
	glUniform3f(getUniLoc(ShadeProg,	"camPos"), eye.x, eye.y, eye.z);
	
	
	glUniform1i(getUniLoc(ShadeProg, "baseTex"), 0);


	glUniform1i(getUniLoc(ShadeProg, "useWalls"), false);
	glUniform1i(getUniLoc(ShadeProg, "wallTex"), 1);
	glUniform3f(getUniLoc(ShadeProg, "wallFreq"),1,1,1);

	glUniform4f(getUniLoc(ShadeProg, "lightPos"),2.0,2.0,2.0,1.0);


	glUniform4f(getUniLoc(ShadeProg, "lightColor"),1.0,1.0,1.0,1.0);

	glUniform3f(getUniLoc(ShadeProg, "offset"),(Pos+neg)/2.0,(Pos+neg)/2.0,(Pos+neg)/2.0);
	
	glPushMatrix();
	glTranslatef(0.5,0.5,0.5);
	glutSolidSphere(250,50,50);
	glPopMatrix();
	

	


	
//	glPushMatrix();
//	glTranslatef(2,2,2);
//	glutSolidSphere(0.1,8,8);
//	glPopMatrix();



	glUniform1i(getUniLoc(ShadeProg, "useWalls"), true);
	
	int face; // negative and positive x, y, z
	int vert;
	for (face = 0; face < 6; face++)
	{
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, baseTexID[face]);
		
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexID[face]); // each face gets its own set of cubemap textures
		
		glBegin(GL_POLYGON);
		glNormal3fv(cubeNormals[face]);
		for (vert = 0; vert < 4; ++vert)
		{
			glMultiTexCoord2fvARB(GL_TEXTURE0_ARB, texCoords[vert]);
			glMultiTexCoord3fvARB(GL_TEXTURE1_ARB, cubeVerts[face][vert]);
			glVertex3fv(cubeVerts[face][vert]);
		}
		glEnd();
		glDisable(GL_TEXTURE_CUBE_MAP);
	}
	glUseProgram(0);
	glDisable(GL_TEXTURE_2D);

}

/*Display function */
void display() {

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
		//set up the camera
	gluLookAt(eye.x, eye.y, eye.z, la.x, la.y, la.z, up.x, up.y, up.z);




	cerr << "eye <" << eye.x << " " << eye.y << " " << eye.z << '>' << endl;
	cerr << "gaze: <" << la.x-eye.x << " " << la.y-eye.y << " " << la.z-eye.z << '>'<< endl;
	cerr << "up <" << up.x << " " << up.y << " " << up.z << '>' << endl;	
	cerr << "theta = " << theta << " phi = " << phi << '\n';
	
	glPushMatrix();
	
	cout << "using a shader" << endl;
	
	
	DrawCube();


		
	glPopMatrix();
	glPopMatrix();
	glutSwapBuffers();

}

#define PI 3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679


/* ASWD camera controls */
void keyboard(unsigned char key, int x, int y) {

	float delta = 0.2;
	Vector3 gaze,left;
	gaze.x = la.x - eye.x;
	gaze.y = la.y - eye.y;
	gaze.z = la.z - eye.z;
//	up = normalize(Vector3(up.x + gaze.x, up.y+gaze.y, up.z+gaze.z)) ;

	gaze = normalize(gaze);

	left = cross(up, gaze);
	cerr << endl;
	switch(key) {
	  case 'a':
		eye.x += delta*left.x;
		eye.y += delta*left.y;
		eye.z += delta*left.z;
//		la.x += delta*left.x;
//		la.y += delta*left.y;
//		la.z += delta*left.z;
		break;
	  case 'd':
		eye.x -= delta*left.x;
		eye.y -= delta*left.y;
		eye.z -= delta*left.z;
//		la.x -= delta*left.x;
//		la.y -= delta*left.y;
//		la.z -= delta*left.z;
		break;
	  case 's':
		eye.x -= delta*gaze.x;
		eye.y -= delta*gaze.y;
		eye.z -= delta*gaze.z;
//		la.x -= delta*gaze.x;
//		la.y -= delta*gaze.y;
//		la.z -= delta*gaze.z;
		break;
	   case 'w':
		eye.x += delta*gaze.x;
		eye.y += delta*gaze.y;
		eye.z += delta*gaze.z;
//		la.x += delta*gaze.x;
//		la.y += delta*gaze.y;
//		la.z += delta*gaze.z;
		break;

	  case 'q':
	  	theta = (theta + 15);
	  	if (theta > 180)
	  		theta -= 360;
	  	break;
	  case 'e':
	  	theta = (theta - 15);
	  	if (theta < -180)
	  		theta += 360;
	  	break;
	  	
	  case 'r':
	  	if (phi < -85)
	  	{
	  		phi = -85;
  		}
  		else
  		{
		  	phi = min(phi + 5, 89.99997f); // heh, that rounds to 90.0 in printing, yet doesn't break the camera.
	  	}
	  	break;
	  case 'f':
	  	if (phi > 85)
	  	{
	  		phi = 85;
  		}
  		else
  		{
		  	phi = max(phi - 5, -89.99997f); 
	  	}
	  	break;
	  	
		
//	  case 'x':
//		  exit( EXIT_SUCCESS );
//		  break;
//	  case 'l':
//		  gbMultitexturing +=1;
//		  gbMultitexturing = gbMultitexturing%3;
//		  printf("gbmult %d\n", gbMultitexturing);
//		  break;
//	  default:
//		  break;
	}



	left = normalize(left);
	
//	up = Vector3(0,1,0);

	gaze = normalize(gaze);

//	float upphi = phi;
	la.x = eye.x + 2 * cos((90-theta) * PI / 180) * sin((90-phi) * PI / 180);
	la.y = eye.y + 2 * 						        cos((90-phi) * PI / 180);
	la.z = eye.z + 2 * sin((90-theta) * PI / 180) * sin((90-phi) * PI / 180);

	gaze.x = la.x - eye.x;
	gaze.y = la.y - eye.y;
	gaze.z = la.z - eye.z;

	up = Vector3(0,1,0);
//	up.x = cos((90-theta) * PI / 180) * sin((upphi) * PI / 180);
//	up.y = 						        cos((upphi) * PI / 180);
//	up.z = sin((90-theta) * PI / 180) * sin((upphi) * PI / 180);
//	
	
	cerr << endl;
	
	glutPostRedisplay();
	
}

void reshape(int w, int h) {
	GW = w;
	GH = h;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(60.0, (float)w/h, 0.1, 500);
	glMatrixMode(GL_MODELVIEW);
	glViewport(0, 0, w, h);

	glutPostRedisplay();
}


int main( int argc, char** argv ) {

	chdir(findMyPath(argv[0]));
	
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH|GLUT_ALPHA);
	glutInitWindowSize(1024, 1024); 
	glutInitWindowPosition(100, 100);
	glutCreateWindow("Interior Mapping");

	glClearColor(1.0, 1.0, 1.0, 1.0);

	/*register glut call backs */
	glutDisplayFunc( display );
	glutReshapeFunc( reshape );
	glutKeyboardFunc(keyboard);
	
	/* projection/modelview set up */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, 1.0, 0.1, 500);
	glMatrixMode(GL_MODELVIEW);
	glViewport(0, 0, 500, 500); //image to pixel

    /*initialize camera stuff */
	GW=GH = 500;
	phi = 0.0; theta = 90.0;	

	eye.x = -4; eye.y = (Pos+neg)/2.0; eye.z = (Pos+neg)/2.0;
	
	up.x = 0.0; up.y = 1; up.z = 0.0;
	
	la.x = (Pos+neg)/2.0; la.y = (Pos+neg)/2.0; la.z = (Pos+neg)/2.0;

	
	keyboard(0,0,0);




	baseTex = new TGA*[N_CUBES];
	baseTexID = new GLuint[N_CUBES];

	cubeTex = new TGA**[N_CUBES];
	cubeTexID = new GLuint[N_CUBES]; 
	
	int set,face;
	for (set = 0; set < N_CUBES; set++) // 6 sets of walls
	{

		glGenTextures(1, &baseTexID[set]);

		baseTex[set] = readTga("textures/" + wallnames[set] + '/' + EXTERIOR_NAME + ".tga"); // result: "textures/Pond/negz.tga"

		glBindTexture(GL_TEXTURE_2D, baseTexID[set]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, baseTex[set]->width, baseTex[set]->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, baseTex[set]->pixels);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		
		glGenTextures(1, &cubeTexID[set]); 
		glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexID[set]); 
		cubeTex[set] = new TGA*[6];
		
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_REPLACE);
	
		for (face = 0; face < 6; face++)
		{
			cubeTex[set][face] = readTga("textures/" + wallnames[set] + '/' + wall_axes[face] + ".tga"); // result: "textures/Pond/negz.tga"
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X+face, 0, GL_RGBA, cubeTex[set][face]->width, cubeTex[set][face]->height,0, GL_RGBA, GL_UNSIGNED_BYTE, cubeTex[set][face]->pixels);
		}
	}

	

	
	Initialize ();
	
	/*set up shader after we've loaded in the textures*/
	getGLversion();
	if (!InstallShader(textFileRead("VLight.glsl"),textFileRead("FLight.glsl"))) {
		printf("\nError installing shader!\n");
		return 0;
	}

	glutMainLoop();
}


