//from CSC 471 - texture mapping lab
//ZJ WOOD

#include <stdlib.h>
#include <GL/glut.h>
#include "GL/glext.h"
#include "GL/glu.h"
#include <stdio.h>
#include <iostream>
#include <math.h>
#include <assert.h>
#include <vector>
#include "tex.h"

RGB* pixel;
Image *TextureImage;
using namespace std;

//code to create a checker board texture...
void makeCheckerBoard ( int nRows, int nCols )
{
  pixel = new RGB[nRows * nCols];
  int c;
  
  long count = 0;
  for ( int i=0; i < nRows; i++ ) {
    for ( int j=0; j < nCols; j++ ) {
      count = j*nCols +i;
      c = (((i/8) + (j/8)) %2) * 255;
      pixel[count].r = c;
      pixel[count].g = c;
      pixel[count].b = c;
      
    }
  }
}


//this is old init_tex, becareful if using with multitexturing as it names the texure 0
// old initialization for texture mapping
void init_tex() {
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, 0);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 64, 64, 0, GL_RGB, GL_UNSIGNED_BYTE, pixel);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}


//routines to load in a bmp files - must be 2^nx2^m and a 24bit bmp
GLvoid LoadTexture(char* image_file, int texID) { 
  
  TextureImage = (Image *) malloc(sizeof(Image));
  if (TextureImage == NULL) {
    printf("Error allocating space for image");
    exit(1);
  }
//  cout << "trying to load " << image_file << endl;
  if (!ImageLoad(image_file, TextureImage)) {
    exit(1);
  }  
  /*  2d texture, level of detail 0 (normal), 3 components (red, green, blue),            */
  /*  x size from image, y size from image,                                              */    
  /*  border 0 (normal), rgb color data, unsigned byte data, data  */ 
  glBindTexture(GL_TEXTURE_2D, texID);
  glTexImage2D(GL_TEXTURE_2D, 0, 3,
    TextureImage->sizeX, TextureImage->sizeY,
    0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage->data);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); /*  cheap scaling when image bigger than texture */    
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); /*  cheap scaling when image smalled than texture*/
  
}


/* BMP file loader loads a 24-bit bmp file only */

/*
* getint and getshort are help functions to load the bitmap byte by byte
*/
static unsigned int getint(FILE *fp) {
  int c, c1, c2, c3;
  
  /*  get 4 bytes */ 
  c = getc(fp);  
  c1 = getc(fp);  
  c2 = getc(fp);  
  c3 = getc(fp);
  
  return ((unsigned int) c) +   
    (((unsigned int) c1) << 8) + 
    (((unsigned int) c2) << 16) +
    (((unsigned int) c3) << 24);
}

static unsigned int getshort(FILE *fp){
  int c, c1;
  
  /* get 2 bytes*/
  c = getc(fp);  
  c1 = getc(fp);
  
  return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

/*  quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  */

int ImageLoad(char *filename, Image *image) {
  FILE *file;
  unsigned long size;                 /*  size of the image in bytes. */
  unsigned long i;                    /*  standard counter. */
  unsigned short int planes;          /*  number of planes in image (must be 1)  */
  unsigned short int bpp;             /*  number of bits per pixel (must be 24) */
  char temp;                          /*  used to convert bgr to rgb color. */
  
  /*  make sure the file is there. */
  if ((file = fopen(filename, "rb"))==NULL) {
    printf("File Not Found : %s\n",filename);
    return 0;
  }
  
  /*  seek through the bmp header, up to the width height: */
  fseek(file, 18, SEEK_CUR);
  
  /*  No 100% errorchecking anymore!!! */
  
  /*  read the width */
  image->sizeX = getint (file);
  
  /*  read the height */ 
  image->sizeY = getint (file);
  
  /*  calculate the size (assuming 24 bits or 3 bytes per pixel). */
  size = image->sizeX * image->sizeY * 3;
  
  /*  read the planes */    
  planes = getshort(file);
  if (planes != 1) {
    printf("Planes from %s is not 1: %u\n", filename, planes);
    return 0;
  }
  
  /*  read the bpp */    
  bpp = getshort(file);
  if (bpp != 24) {
    printf("Bpp from %s is not 24: %u\n", filename, bpp);
    return 0;
  }
  
  /*  seek past the rest of the bitmap header. */
  fseek(file, 24, SEEK_CUR);
  
  /*  read the data.  */
  image->data = (char *) malloc(size);
  if (image->data == NULL) {
    printf("Error allocating memory for color-corrected image data");
    return 0; 
  }
  
  if ((i = fread(image->data, size, 1, file)) != 1) {
    printf("Error reading image data from %s.\n", filename);
    return 0;
  }
  
  for (i=0;i<size;i+=3) { /*  reverse all of the colors. (bgr -> rgb) */
    temp = image->data[i];
    image->data[i] = image->data[i+2];
    image->data[i+2] = temp;
  }
  
  fclose(file); /* Close the file and release the filedes */
  
  /*  we're done. */
  return 1;
}





void MergeBytes(PIXEL *,unsigned char *,int);



TGA * readTga(const std::string filenamestr)
{
	const char* filename = filenamestr.c_str();
	cerr << filename << endl;
   int n=0,i,j;
   int bytesTOread,skipover = 0;
   unsigned char p[5];
   FILE *fptr;
   
   TGA *header = (TGA*) malloc(sizeof(TGA));;
   if (!header)
   {
   		perror("malloc");
   		exit(-1);
	}
	
   PIXEL *pixels;

   /* Open the file */

   if ((fptr = fopen(filename,"r")) == NULL) {
      perror("File open failed");
      exit(-1);
   }


   /* Display the header fields */
   header->idlength = fgetc(fptr);
//   fprintf(stderr,"ID length:         %d\n",header->idlength);
   header->colourmaptype = fgetc(fptr);
//   fprintf(stderr,"Colourmap type:    %d\n",header->colourmaptype);
   header->datatypecode = fgetc(fptr);
//   fprintf(stderr,"Image type:        %d\n",header->datatypecode);
   fread(&header->colourmaporigin,2,1,fptr);
//   fprintf(stderr,"Colour map offset: %d\n",header->colourmaporigin);
   fread(&header->colourmaplength,2,1,fptr);
//   fprintf(stderr,"Colour map length: %d\n",header->colourmaplength); 
   header->colourmapdepth = fgetc(fptr);
//   fprintf(stderr,"Colour map depth:  %d\n",header->colourmapdepth);
   fread(&header->x_origin,2,1,fptr);
//   fprintf(stderr,"X origin:          %d\n",header->x_origin);
   fread(&header->y_origin,2,1,fptr);
//   fprintf(stderr,"Y origin:          %d\n",header->y_origin);
   fread(&header->width,2,1,fptr);
//   fprintf(stderr,"Width:             %d\n",header->width);
   fread(&header->height,2,1,fptr);
//   fprintf(stderr,"Height:            %d\n",header->height);
   header->bitsperpixel = fgetc(fptr);
//   fprintf(stderr,"Bits per pixel:    %d\n",header->bitsperpixel);
   header->imagedescriptor = fgetc(fptr);
//   fprintf(stderr,"Descriptor:        %d\n",header->imagedescriptor);

   /* Allocate space for the image */
   if ((pixels = (PIXEL*)malloc(header->width*header->height*sizeof(PIXEL))) == NULL) {
      fprintf(stderr,"malloc of image failed\n");
      exit(-1);
   }
   for (i=0;i<header->width*header->height;i++) {
      pixels[i].r = 0;
      pixels[i].g = 0;
      pixels[i].b = 0;
      pixels[i].a = 0;
   }

   /* What can we handle */
   if (header->datatypecode != 2 && header->datatypecode != 10) {
      fprintf(stderr,"Can only handle image type 2 and 10\n");
      exit(-1);
   }
   if (header->bitsperpixel != 16 && 
       header->bitsperpixel != 24 && header->bitsperpixel != 32) {
      fprintf(stderr,"Can only handle pixel depths of 16, 24, and 32\n");
      exit(-1);
   }
   if (header->colourmaptype != 0 && header->colourmaptype != 1) {
      fprintf(stderr,"Can only handle colour map types of 0 and 1\n");
      exit(-1);
   }

   /* Skip over unnecessary stuff */
   skipover += header->idlength;
   skipover += header->colourmaptype * header->colourmaplength;
//   fprintf(stderr,"Skip over %d bytes\n",skipover);
   fseek(fptr,skipover,SEEK_CUR);

   /* Read the image */
   bytesTOread = header->bitsperpixel / 8;
   while (n < header->width * header->height) {
      if (header->datatypecode == 2) {                     /* Uncompressed */
         if (fread(p,1,bytesTOread,fptr) != bytesTOread) {
            fprintf(stderr,"Unexpected end of file at pixel %d\n",i);
            exit(-1);
         }
         MergeBytes(&(pixels[n]),p,bytesTOread);
         n++;
      } else if (header->datatypecode == 10) {             /* Compressed */
         if (fread(p,1,bytesTOread+1,fptr) != bytesTOread+1) {
            fprintf(stderr,"Unexpected end of file at pixel %d\n",i);
            exit(-1);
         }
         j = p[0] & 0x7f;
         MergeBytes(&(pixels[n]),&(p[1]),bytesTOread);
         n++;
         if (p[0] & 0x80) {         /* RLE chunk */
            for (i=0;i<j;i++) {
               MergeBytes(&(pixels[n]),&(p[1]),bytesTOread);
               n++;
            }
         } else {                   /* Normal chunk */
            for (i=0;i<j;i++) {
               if (fread(p,1,bytesTOread,fptr) != bytesTOread) {
                  fprintf(stderr,"Unexpected end of file at pixel %d\n",i);
                  exit(-1);
               }
               MergeBytes(&(pixels[n]),p,bytesTOread);
               n++;
            }
         }
      }
   }
   fclose(fptr);
   header->pixels = pixels;
	return header;

}

void MergeBytes(PIXEL *pixel,unsigned char *p,int bytes)
{
   if (bytes == 4) {
      pixel->r = p[2];
      pixel->g = p[1];
      pixel->b = p[0];
      pixel->a = p[3];
   } else if (bytes == 3) {
      pixel->r = p[2];
      pixel->g = p[1];
      pixel->b = p[0];
      pixel->a = ~0; // needs to default to OPAQUE, GRR! (was '0')
   } else if (bytes == 2) {
      pixel->r = (p[1] & 0x7c) << 1;
      pixel->g = ((p[1] & 0x03) << 6) | ((p[0] & 0xe0) >> 2);
      pixel->b = (p[0] & 0x1f) << 3;
      pixel->a = (p[1] & 0x80);
   }
}


